"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var fs = require("fs");
var Path = require("path");
var crypto = require("crypto");
var getStream = require("get-stream");
var request = require("request");
var requestPromise = require("request-promise");
var tar = require("tar");
var unzipper = require("unzipper");
var cmdShim = require("cmd-shim");
var util_1 = require("util");
var child_process_1 = require("child_process");
/** return the sha256 of a file as a hex-formatted string */
function sha256OfFile(path) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var input, cipher;
        return tslib_1.__generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    input = fs.createReadStream(path);
                    cipher = crypto.createHash('sha256');
                    return [4 /*yield*/, getStream(input.pipe(cipher), { encoding: 'hex' })];
                case 1: return [2 /*return*/, (_c.sent()).toUpperCase()];
            }
        });
    });
}
exports.sha256OfFile = sha256OfFile;
/** download a URL and save it to a file */
function downloadUrlToFile(url, path, requestOpts) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var fileStream;
        return tslib_1.__generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    fileStream = fs.createWriteStream(path);
                    return [4 /*yield*/, new Promise(function (res, rej) {
                            request(url, requestOpts).pipe(fileStream).on('close', function () {
                                res();
                            });
                        })];
                case 1:
                    _c.sent();
                    return [2 /*return*/];
            }
        });
    });
}
exports.downloadUrlToFile = downloadUrlToFile;
/** download a URL and return it as a string */
function downloadUrlAsString(url, requestOpts) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        return tslib_1.__generator(this, function (_c) {
            switch (_c.label) {
                case 0: return [4 /*yield*/, requestPromise(url, requestOpts)];
                case 1: return [2 /*return*/, _c.sent()];
            }
        });
    });
}
exports.downloadUrlAsString = downloadUrlAsString;
/** download a URL and return it as parsed JSON */
function downloadUrlAsJson(url, requestOpts) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var _c, _d;
        return tslib_1.__generator(this, function (_e) {
            switch (_e.label) {
                case 0:
                    _d = (_c = JSON).parse;
                    return [4 /*yield*/, downloadUrlAsString(url, requestOpts)];
                case 1: return [2 /*return*/, _d.apply(_c, [_e.sent()])];
            }
        });
    });
}
exports.downloadUrlAsJson = downloadUrlAsJson;
function readFileWithDefault(path, defaultContent) {
    try {
        return fs.readFileSync(path, 'utf8');
    }
    catch (e) {
        if (!fs.existsSync(path))
            return defaultContent;
        throw e;
    }
}
exports.readFileWithDefault = readFileWithDefault;
/** Read a JSON file from disk, automatically parsing it, returning default value if the file doesn't exist. */
function readJsonFileWithDefault(path, defaultContent) {
    try {
        return JSON.parse(fs.readFileSync(path, 'utf8'));
    }
    catch (e) {
        if (!fs.existsSync(path))
            return defaultContent;
        throw e;
    }
}
exports.readJsonFileWithDefault = readJsonFileWithDefault;
/** Write a JSON file to disk by stringify-ing the value.  Writes UTF-8 encoding. */
function writeJsonFile(path, value, indent) {
    if (indent === void 0) { indent = true; }
    if (indent === true)
        indent = '    ';
    fs.writeFileSync(path, JSON.stringify(value, null, indent));
}
exports.writeJsonFile = writeJsonFile;
function patchJsonFile(path, _a, _b) {
    var _c;
    var _d = [_a, _b], _e = _d[0], opts = _e === void 0 ? {} : _e, callback = _d[1];
    if (!callback)
        _c = [{}, _a], opts = _c[0], callback = _c[1];
    var _f = opts.defaultValue, defaultValue = _f === void 0 ? null : _f, _g = opts.indent, indent = _g === void 0 ? false : _g;
    var value = readJsonFileWithDefault(path, defaultValue);
    var newValue = callback(value);
    if (newValue === undefined)
        newValue = value;
    writeJsonFile(path, newValue, indent);
}
exports.patchJsonFile = patchJsonFile;
function extractArchive(type, archivePath, destination) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        return tslib_1.__generator(this, function (_c) {
            switch (type) {
                case '.tar.gz':
                    return [2 /*return*/, extractTarFile(archivePath, destination)];
                    break;
                case '.zip':
                    return [2 /*return*/, extractZipFile(archivePath, destination)];
                    break;
                default:
                    throw new Error('Unsupported archive type: ' + type);
            }
            return [2 /*return*/];
        });
    });
}
exports.extractArchive = extractArchive;
/** Extract a tar.gz file into a destination directory. */
function extractTarFile(tarPath, destination) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        return tslib_1.__generator(this, function (_c) {
            tar.x({
                cwd: destination,
                file: tarPath,
                sync: true
            });
            return [2 /*return*/];
        });
    });
}
exports.extractTarFile = extractTarFile;
/** Extract a zip file into a destination directory. */
function extractZipFile(zipPath, destination) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        return tslib_1.__generator(this, function (_c) {
            switch (_c.label) {
                case 0: return [4 /*yield*/, finished(fs.createReadStream(zipPath).pipe(unzipper.Extract({ path: destination })))];
                case 1: return [2 /*return*/, _c.sent()];
            }
        });
    });
}
exports.extractZipFile = extractZipFile;
function finished(stream, waitForEvent) {
    if (waitForEvent === void 0) { waitForEvent = 'finish'; }
    return new Promise(function (res, rej) {
        stream.on(waitForEvent, function () {
            res();
        });
        stream.on('error', rej);
    });
}
/** spawn a process on PATH, get the full stdout as a string.  Throw if process returns non-zero status or anything else goes wrong. */
function getStdout(commandAndArgs) {
    var result = child_process_1.spawnSync(commandAndArgs[0], commandAndArgs.slice(1));
    if (result.status !== 0)
        throw new Error("process returned non-zero status: " + result.status);
    return result.stdout;
}
exports.getStdout = getStdout;
/** Returns npm prefix, verbatim (no path normalization, straight from the `npm` command) */
function getNpmPrefix() {
    // TODO memoize this
    return getStdout(['npm', 'config', 'get', 'prefix']);
}
exports.getNpmPrefix = getNpmPrefix;
/** get absolute path to the `bin` or `.bin` directory into which npm will install binaries (either symlinks or .cmd stubs) */
function getNpmBinDirectory() {
    if (exports.isGlobalInstall) {
        switch (process.platform) {
            case 'win32':
                return Path.normalize(getNpmPrefix());
                break;
            case 'linux':
            case 'darwin':
                return Path.resolve(getNpmPrefix(), 'bin');
                break;
            default:
                throw new Error("Unsupported: global installation on " + process.platform);
        }
    }
    // Local installation: find the local node_modules/.bin
    else {
        if (process.env.npm_lifecycle_event) {
            return Path.resolve(__dirname, '../../.bin');
        }
        else {
            // This only happens when we're testing
            return Path.resolve(__dirname, '../test-bin');
        }
    }
}
exports.getNpmBinDirectory = getNpmBinDirectory;
/**
 * Return absolute, normalized path to the shim that NPM would generate for a package bin script
 */
function getNpmBinShimPath(name) {
    return Path.normalize(Path.join(getNpmBinDirectory(), name + (process.platform === 'win32' ? '.cmd' : '')));
}
exports.getNpmBinShimPath = getNpmBinShimPath;
/**
 * Return absolute, normalized path to npm's global node_modules directory, where modules are installed globally
 */
function getNpmGlobalNodeModules() {
    switch (process.platform) {
        case 'win32':
            return Path.resolve(getNpmPrefix(), 'node_modules');
            break;
        case 'linux':
        case 'darwin':
            return Path.resolve(getNpmPrefix(), 'lib', 'node_modules');
            break;
        default:
            throw new Error("Unsupported: global installation on " + process.platform);
    }
}
exports.getNpmGlobalNodeModules = getNpmGlobalNodeModules;
/** True if this is an `npm install --global`, false if it's a local install */
exports.isGlobalInstall = !!process.env.npm_config_global;
/** Return true if paths point to the same file or directory, taking OS differences into account */
function pathsEquivalent(path1, path2) {
    path1 === fullyNormalizePath(path1);
    path2 === fullyNormalizePath(path2);
    return path1 === path2;
}
exports.pathsEquivalent = pathsEquivalent;
/** More aggressive path normalization. Convert to lowercase on Windows and strip trailing path separators */
function fullyNormalizePath(path) {
    path = Path.normalize(path);
    if (process.platform === 'win32') {
        path = path.toLowerCase();
    }
    while (path[path.length - 1] === Path.sep) {
        path = path.slice(0, -1);
    }
    return path;
}
exports.fullyNormalizePath = fullyNormalizePath;
function createSymlinkTo(linkPath, targetPath, log) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        return tslib_1.__generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    try {
                        fs.unlinkSync(linkPath);
                    }
                    catch (e) {
                        if (e.code !== 'ENOENT')
                            throw e;
                    }
                    if (!(process.platform === 'win32')) return [3 /*break*/, 2];
                    log("Creating .cmd shim from " + linkPath + " to " + targetPath + "...");
                    return [4 /*yield*/, util_1.promisify(cmdShim)(targetPath, linkPath.replace(/\.cmd$/, ''))];
                case 1:
                    _c.sent();
                    return [3 /*break*/, 3];
                case 2:
                    log("Symlinking from " + linkPath + " to " + targetPath + "...");
                    fs.symlinkSync(targetPath, linkPath);
                    _c.label = 3;
                case 3: return [2 /*return*/];
            }
        });
    });
}
exports.createSymlinkTo = createSymlinkTo;
